from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np

from ..grids import ComplexGrid, RealGrid

if TYPE_CHECKING:
    from zlug.file.proxies import ObjectProxy


class ComplexGridH5:
    @staticmethod
    def read(path: str, **kwargs) -> ComplexGrid:
        raise NotImplementedError

    @staticmethod
    def write(path: str, data: ComplexGrid, **kwargs) -> None:
        raise NotImplementedError


class ComplexGridNPZ:
    @staticmethod
    def read(path: str, **kwargs) -> ComplexGrid:
        with open(path, 'rb') as f:
            dat = np.load(f) 
            psi = dat['psi']
            shape = psi.shape
            grid = ComplexGrid(shape)
            grid.psi[...] = psi
        return grid

    @staticmethod
    def write(path: str, data: ComplexGrid, **kwargs) -> None:
        with open(path, 'wb') as f:
            np.savez(f, psi=data.psi)



from __future__ import annotations
from typing import Literal, Type, Union, Sequence
import numpy as np
import numpy.typing as npt
from enum import Enum


class FloatingPointPrecision(Enum):
    SINGLE = 'SINGLE'
    DOUBLE = 'DOUBLE'
    LONGDOUBLE = 'LONGDOUBLE'

    def __str__(self) -> str:
        return self.name


PrecisionStr = Literal[
    'SINGLE', 'DOUBLE', 'LONGDOUBLE',
    'single', 'double', 'longdouble',
]

_real_dtype = {
    'SINGLE': np.single,
    'DOUBLE': np.double,
    'LONGDOUBLE': np.longdouble,
}

_complex_dtype = {
    'SINGLE': np.csingle,
    'DOUBLE': np.cdouble,
    'LONGDOUBLE': np.clongdouble,
}

PrecisionLike = Union[FloatingPointError, PrecisionStr]

def get_real_dtype(
        precision: PrecisionLike
    ) -> Type[np.floating]:
    return _real_dtype[str(precision).upper()]


def get_complex_dtype(
        precision: PrecisionLike
    ) -> Type[np.complexfloating]:
    return _complex_dtype[str(precision).upper()]



NPFloat = Union[np.floating, np.float_]
'single, double, longdouble, float_'

FloatLike = Union[NPFloat, float]
'single, double, longdouble, float_, float'

NPComplex = Union[np.complexfloating, np.complex_]
'csingle, cdouble, clongdouble, complex_'

ComplexLike = Union[NPComplex, complex]
'csingle, cdouble, clongdouble, complex_, complex'


SizeLike = Union[Sequence[FloatLike], npt.NDArray[np.floating], npt.NDArray[np.float_]]






from __future__ import annotations
import numpy as np


from ..core.fourier import generate_xk

from ..grids import ComplexGrid
from ._base import Field


class ComplexField(Field[np.complexfloating], ComplexGrid):
    """
    A field is a grid with length scales. While the grid shape is fixed for a:
    particular instance, the system size (lengths in each dimension) can be
    changed with set_size().

    """
    def _update_coordinate_vars(self) -> None:

        R, K, DR, DK = [], [], [], []

        self._R.setflags(write=True)
        self._K.setflags(write=True)
        self._dR.setflags(write=True)
        self._dK.setflags(write=True)

        for i in range(self.rank):
            r, k, dr, dk = generate_xk(self.size[i], self.shape[i], precision=self.precision)
            R.append(r)
            K.append(k)
            DR.append(dr)
            DK.append(dk)

        self._R[...] = np.meshgrid(*R, indexing='ij')
        self._K[...] = np.meshgrid(*K, indexing='ij')
        self._dR[...] = np.array(DR)
        self._dK[...] = np.array(DK)

        self._R.setflags(write=False)
        self._K.setflags(write=False)
        self._dR.setflags(write=False)
        self._dK.setflags(write=False)





from __future__ import annotations
import numpy as np
from ..fields import RealField2D
from .. import core


class RealField2DNPZ:
    """
    RealField2D to NPZ
    """
    @staticmethod
    def read(path: str, **kwargs) -> RealField2D:
        with open(path, 'rb') as f:
            dat = np.load(f, allow_pickle=False, **kwargs)

            state = dict()
            for key in dat.files:
                state[key] = dat[key]

            state = core.scalarize(state)

            psi = state['psi']
            Lx = state['Lx']
            Ly = state['Ly']
            Nx = psi.shape[0]
            Ny = psi.shape[1]

            field = RealField2D(Lx, Ly, Nx, Ny)
            field.psi[...] = psi
        return field

    @staticmethod
    def write(path: str, data: RealField2D, **kwargs) -> None:
        assert isinstance(data, RealField2D)
        with open(path, 'wb') as f:
            np.savez(f,
                psi=data.psi.copy(),
                Lx=data.lx,
                Ly=data.ly, **kwargs
            )




from __future__ import annotations
from typing import Tuple, final

import numpy as np
from pyfftw import pyfftw
import warnings

from ..core import get_complex_dtype, get_real_dtype
from ._base import Grid


class RealGrid(Grid[np.floating]):
    """
    A RealGridND object is a real array of shape (d1, d2, .., dN) equipped
    with fourier transform. No length scales are associated with the grid.

    The generic type variables are the dtypes of self.psi & self.psi_k.
    """
    @final
    def _init_grid_data(self, shape: Tuple[int, ...]):

        last_axis = self._fft_axes[-1]

        if shape[last_axis] % 2 != 0:
            warnings.warn('odd resolution on the last axis will be automatically made even for RFFT')
            shape_ = list(shape)
            shape_[last_axis] += 1
            shape = tuple(shape_)

        shape_k = list(shape)
        shape_k[last_axis] = shape_k[last_axis]//2 + 1
        shape_k = tuple(shape_k)

        self._psi = pyfftw.zeros_aligned(shape, dtype=get_real_dtype(self._precision))
        self._psi_k = pyfftw.zeros_aligned(shape_k, dtype=get_complex_dtype(self._precision))

    @property
    @final
    def isreal(self): return True

    # def set_psi(self, psi1: Union[float, npt.NDArray[np.floating]]):
    #     if not np.isscalar(psi1):
    #         assert isinstance(psi1, np.ndarray)
    #         if psi1.shape != self.shape:
    #             raise ValueError(f'array has incompatible shape {psi1.shape} with {self.shape}')
    #
    #     if not np.all(np.isreal(psi1)):
    #         raise ValueError(f'array is complex') 
    #     self.psi[...] = psi1

    # def __init__(self,
    #         shape: Tuple[int, ...], *,
    #         precision: PrecisionStr = 'double',
    #         fft_axes: Optional[Tuple[int,...]] = None
    #     ):
    #     """
    #     Parameters: 
    #         shape: tuple of integers (d1, d2, ..., dN),
    #
    #
    #     The last axis in self.fft_axes must be even
    #     (if not it will be made even automatically)
    #     """
    #     self._precision = precision
    #
    #     if fft_axes is None:
    #         self._fft_axes = tuple(np.arange(len(shape)))
    #     else:
    #         self._fft_axes = fft_axes
    #
    #     last_axis = self._fft_axes[-1]
    #
    #     if shape[last_axis] % 2 != 0:
    #         warnings.warn('odd resolution on the last axis will be automatically made even for RFFT')
    #         shape_list = list(shape)
    #         shape_list[last_axis] += 1
    #         shape = tuple(shape_list)
    #
    #     shape_k = list(shape)
    #     shape_k[last_axis] = shape_k[last_axis]//2 + 1
    #     shape_k = tuple(shape_k)
    #
    #     self._psi = pyfftw.zeros_aligned(shape, dtype=get_real_dtype(self._precision))
    #     self._psi_k = pyfftw.zeros_aligned(shape_k, dtype=get_complex_dtype(self._precision))




from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Dict, Type, TypeVar

from .hooks import EvolverHooks, DefaultHooks
from ..misc import context

from ..core import generic

import threading

import time

T = TypeVar('T')


from typing import Generic, Optional


@generic
class Evolver(ABC, Generic[T]):
    '''
    An Evolver handles the evolution of an object. There's no limit on
    the object's type.
    '''
    def __init__(self, subject: T):
        self.subject = subject
        self._continue_flag = True

        # Backward compatibility
        self.run_nonstop = self.run
        
        self.data: Dict[str, Any] = {}
        '''
        Used to store fields used by hooks.
        '''

        self.thread: threading.Thread|None = None

    @abstractmethod
    def step(self):
        '''
        Run a single minimization step.
        '''
        raise NotImplementedError

    def run_steps(self, n_steps: int):
        '''
        Run [n_steps] minimization steps.
        '''
        for i in range(n_steps):
            self.step()
            i += 1 

    def run_multisteps(self, 
            n_steps: int, n_epochs: int, 
            hooks: Optional[EvolverHooks]=None):
        '''
        Run evolution loop 
        '''
        hooks = self.resolve_hooks(hooks)

        self.start()
        hooks.on_multisteps_start(n_steps, n_epochs)
        with context.Context(hooks.multisteps_enter, hooks.multisteps_exit):
            for i in range(n_epochs):
                self.run_steps(n_steps)
                hooks.on_multisteps_step(i)

                if not self.get_continue_flag():
                    break
        hooks.on_multisteps_end()

    def run(self, n_steps: int, hooks: Optional[EvolverHooks]=None):
        '''
        Run evolution loop. 
        hooks.on_nonstop_step() will be called every [n_steps] steps.
        The minimization will be run on a separate thread.
        '''
        hooks = self.resolve_hooks(hooks)

        self._lock = threading.Lock()

        def run_():
            self.start()
            hooks.on_nonstop_start(n_steps)
            with context.Context(hooks.nonstop_enter, hooks.nonstop_exit):
                i = 0
                while True:
                    with self._lock:
                        if not self.get_continue_flag():
                            break
                        self.run_steps(n_steps)
                    hooks.on_nonstop_step(i)
                    i += 1

            hooks.on_nonstop_end()

        self.thread = threading.Thread(target=run_)
        self.thread.start()

        while True:
            try:
                time.sleep(1)
                if not self.get_continue_flag():
                    break
            except KeyboardInterrupt:
                hooks.pre_interrupt()
                
                self._lock.acquire()
                handled = hooks.on_interrupt()
                if self._lock.locked(): self._lock.release()

                if not handled:
                    self.set_continue_flag(False)

                hooks.post_interrupt()

        self.thread.join()

    def get_continue_flag(self) -> bool:
        return self._continue_flag

    def set_continue_flag(self, flag: bool):
        self._continue_flag = flag

    def resolve_hooks(self, hooks: Optional[EvolverHooks]) -> EvolverHooks:
        '''
        Each Evolver instance is also an instance of EvolverHooks. By default
        none of the hook methods are implemented.
        '''
        if hooks is None:
            _hooks = DefaultHooks()
        else:
            _hooks = hooks

        _hooks.bind(self)
        return _hooks

    def start(self) -> None:
        '''
        A hook method that is called before evolution
        '''
        self.data.clear()


    def wait(self):
        '''
            Wait till the evolution thread ends
        '''
        if self.thread is not None:
            if self._lock.locked():
                self._lock.release()
            self.thread.join()







raise ImportError('Module is outdated and deprecated')
from __future__ import annotations
from typing import Literal, overload

import numpy as np

from ..core import scalarize

from ._2d import ComplexField2D, RealField2D


@overload
def load_field(filepath: str) -> RealField2D: ...

@overload
def load_field(filepath: str, is_complex: Literal[False]) -> ComplexField2D: ...

@overload
def load_field(filepath: str, is_complex: Literal[True]) -> RealField2D: ...

@overload
def load_field(filepath: str, is_complex: bool) -> RealField2D | ComplexField2D: ...

def load_field(filepath: str, is_complex: bool=False) -> RealField2D | ComplexField2D:
    state = np.load(filepath, allow_pickle=True)
    
    state1 = dict()
    for key in state.files:
        state1[key] = state[key]
    state1 = scalarize(state1)

    return import_field(state1, is_complex=is_complex)


@overload
def import_field(state: dict) -> RealField2D: ...

@overload
def import_field(state: dict, is_complex: Literal[False]) -> RealField2D: ...

@overload
def import_field(state: dict, is_complex: Literal[True]) -> ComplexField2D: ...

@overload
def import_field(state: dict, is_complex: bool) -> ComplexField2D | RealField2D: ...

def import_field(state: dict, is_complex: bool=False) -> ComplexField2D | RealField2D:
    psi = state['psi']
    Lx = state['Lx']
    Ly = state['Ly']
    Nx = psi.shape[0]
    Ny = psi.shape[1]
    if is_complex: 
        field = ComplexField2D(Lx, Ly, Nx, Ny)
        field.set_psi(psi)
        return field
    else:
        field = RealField2D(Lx, Ly, Nx, Ny)
        field.set_psi(psi)
        return field


class FieldOperationError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message






from __future__ import annotations
from abc import abstractmethod
from textwrap import dedent


from typing import Callable, Literal, Optional, TypeVar

import rich
from rich.box import MINIMAL
from rich.console import Group, RenderableType
from rich.panel import Panel
from rich.progress import Progress
from torusgrid.dynamics.hooks.base import EvolverHooks


import numpy as np

from torusgrid.misc.console import make_progress_bar


T = TypeVar('T')


class EarlyStopping(EvolverHooks[T]):
    '''
    Stops the evolver when self.condition() returns true enough times
    '''
    def __init__(self, 
            patience: float, *,
            cumulative: bool=False,
            period: int=1,
            verbose: bool=True,
            show_progress: bool=True
        ):

        self.cumulative = cumulative
        self.check_period = period
        self.patience = patience
        self._badness = 0

        self.verbose = verbose
        self.show_progress = show_progress

    def get_status(self) -> RenderableType:
        return dedent(f'''\
            Monitoring: {self.get_monitor_target()}
            Status: {self._badness}/{self.patience}'''
        )

    def on_start(self, n_steps: int, n_epochs: Optional[int]):
        self._badness = 0

    def enter_(self):
        if self.show_progress:
            self.panel: Panel = self.evolver.data['__panel__']

            self.pbar = make_progress_bar(text=f'Monitoring {self.get_monitor_target()}', 
                    transient=True)

            self.task = self.pbar.add_task(self.get_monitor_target(), total=self.patience)

            self.panel.renderable = Group(self.panel.renderable, self.pbar)



    def on_step(self, step: int):
        if step % self.check_period == 0:
            if self.condition():

                if self.show_progress:
                    self.pbar.advance(self.task)

                self._badness += 1
            else:
                if not self.cumulative:
                    self._badness = 0
                    self.pbar.update(self.task, completed=0)

            if self._badness >= self.patience:
                if self.verbose:
                    rich.get_console().log('Critierion met, stopping...')
                self.evolver.set_continue_flag(False)

    @abstractmethod
    def condition(self) -> bool: ...


    @abstractmethod
    def get_monitor_target(self) -> str: ...


class DetectSlow(EarlyStopping[T]):
    def __init__(self, 
            target: Callable[[T], float] | str,
            rtol: float, atol: float,
            monotone: Literal['increase', 'decrease', 'ignore'],
            patience: float, *, 
            cumulative: bool = False,
            period: int = 1, 
            verbose: bool = True,
            show_progress: bool = True
            ):
        '''
        Monitor a value (via target), and stop the evolver when the value
        changes sufficiently slowly.

        target can be either a function (T) -> float or a string. If a string
        is provided, evolve.data[target] will be used.
        '''

        super().__init__(patience, 
                cumulative=cumulative,
                period=period,
                verbose=verbose,
                show_progress=show_progress)

        self.rtol = rtol
        self.atol = atol

        self.target = target

        self._val_prev: None|float = None

        self.compare = lambda x, y: np.isclose(x, y, rtol=rtol, atol=atol)

        if monotone == 'increase':
            self.compare = lambda x, y: np.isclose(x, y, rtol=rtol, atol=atol) and x >= y

        if monotone == 'decrease':
            self.compare = lambda x, y: np.isclose(x, y, rtol=rtol, atol=atol) and x <= y

        self.verbose = verbose

        self._target = str(self.target)

    def enter_(self):
        super().enter_()

    def get_monitor_target(self):
        return self._target

    def on_start(self, n_steps: int, n_epochs: Optional[int]):
        self._badness = 0
        self._val_prev = None

    def condition(self):
        if isinstance(self.target, str):
            val: float = self.evolver.data[self.target]
        else:
            val = self.target(self.evolver.subject)

        cond = False

        if self._val_prev is not None:
            cond = self.compare(val, self._val_prev)


        self._val_prev = val

        return cond






from __future__ import annotations


from typing import TYPE_CHECKING, Callable, Dict, Literal, TypeVar
from torusgrid.dynamics.hooks.base import EvolverHooks


if TYPE_CHECKING:
    from torusgrid.dynamics.base import Evolver



T = TypeVar('T')

class MonitorValues(EvolverHooks[T]):
    def __init__(self, 
            target: Dict[str, Callable[[Evolver[T]], float|None]] | 
                    Callable[[Evolver[T]], Dict[str, float|None]],
            *, 
            period: int=1) -> None:
        super().__init__()

        self.period = period
        self.target = target

        if isinstance(self.target, dict): 
            def record_():
                for key, func in self.target.items():
                    self.evolver.data[key] = func(self.evolver)
        else:
            def record_():
                values = self.target(self.evolver) # type: ignore
                self.evolver.data.update(values)

        self.record_values = record_

    def on_step(self, step: int):
        if step % self.period == 0:
            self.record_values()






raise ImportError('torusgrid.callbacks is deprecated. Use torusgrid.dynamics.hooks instead')

from __future__ import annotations
from enum import Enum

import numpy as np

from michael960lib.common import overrides, IllegalActionError, ModifyingReadOnlyObjectError

import warnings

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .dynamics import FancyEvolver



class EvolverCallBack:
    def __init__(self):
        pass

    def on_call(self, ev: FancyEvolver, sf: FieldStateFunction):
        raise NotImplementedError

    def stop_evolver(self, ev: FancyEvolver):
        ev.set_continue_flag(False)

    # for reusing the callback obj 
    def reinit(self):
        raise NotImplementedError


class EvolverMonitor(Enum):
    DIFFERENCE = 0
    OSCILLATION = 1
    

class EarlyStopping(EvolverCallBack):
    def __init__(self, patience: int, cumulative=False):
        self.p = 0
        self.patience = patience
        self.cumulative = cumulative

    @overrides(EvolverCallBack)
    def on_call(self, ev: FancyEvolver, sf: FieldStateFunction):
        #ev.info['callback'] = f'{self.p}/{self.patience}'

        if self.condition(ev, sf):
            self.p += 1
        else:
            if not self.cumulative:
                self.p = 0

        if self.p >= self.patience:
            self.stop_evolver(ev)

    def condition(self, ev: FancyEvolver, sf: FieldStateFunction) -> bool:
        return False

    @overrides(EvolverCallBack)
    def reinit(self):
        self.p = 0


class MonitoredEarlyStopping(EarlyStopping):
    def __init__(self, monitor: EvolverMonitor, target_key,
            patience: int, cumulative: bool=False,
            rtol: float=0, atol: float=0):
        super().__init__(patience, cumulative)
        self.monitor = monitor
        self.val = 1e16
        self.val_prev = 1e16
        self.val_prev2 = 1e16

        self.rtol = rtol
        self.atol = atol
        self.target_key = target_key

    @overrides(EarlyStopping)
    def on_call(self, ev: FancyEvolver, sf: FieldStateFunction):
        self.val_prev2 = self.val_prev
        self.val_prev = self.val 
        self.val = sf.get_item(self.target_key)
        super().on_call(ev, sf)

    @overrides(EarlyStopping)
    def condition(self, ev: FancyEvolver, sf: FieldStateFunction) -> bool:
        if self.monitor is EvolverMonitor.DIFFERENCE:
            return np.isclose(self.val, self.val_prev, self.rtol, self.atol)

        if self.monitor is EvolverMonitor.OSCILLATION:
            return (self.val - self.val_prev) * (self.val_prev - self.val_prev2) < 0

        return False

    @overrides(EvolverCallBack)
    def reinit(self):
        super().reinit()
        self.val = 1e16
        self.val_prev = 1e16
        self.val_prev2 = 1e16









from typing import Optional, TypeVar
from torusgrid.dynamics.hooks.base import EvolverHooks
import time

from ...misc import console
import rich


T = TypeVar('T')

class DefaultHooks(EvolverHooks[T]):
    def on_start(self, n_steps: int, n_epochs: Optional[int]):
        self.pbar = console.make_progress_bar()
        self.task = self.pbar.add_task('', total=n_epochs, start=(n_epochs is not None))
        
    def enter_(self):
        self.pbar.__enter__()

    def exit_(self, *args):
        time.sleep(0.1)
        self.pbar.__exit__(*args)

    def on_step(self, step: int):
        self.pbar.advance(self.task)

    def pre_interrupt(self):
        rich.get_console().log('Interrupted')
    



from __future__ import annotations

from abc import abstractmethod
from typing import Tuple, TypeVar

from ...core import FloatLike
from ...grids import Grid
import numpy.typing as npt

from .temporal import TemporalEvolver


T = TypeVar('T', bound=Grid)

class SecondOrderRK4(TemporalEvolver[T]):
    '''
    Performs RK4 on a ComplexGrid object to solve a PDE that is
    second-order in time
    '''
    def __init__(self, grid: T, dt: FloatLike):
        super().__init__(grid, dt)
    
        self.grid = self.subject
        '''
        Here self.grid == self.subject
        '''
        
        self.grid_tmp = self.grid.copy()
        self.grid_tmp.initialize_fft()

        self.dgrid = self.grid.copy()
        self.dgrid.set_psi(0)
        self.dgrid.initialize_fft()

        self.dgrid_tmp = self.dgrid.copy()
        self.dgrid_tmp.initialize_fft()


    @abstractmethod
    def psi_dot(self) -> Tuple[npt.NDArray, npt.NDArray]:
        '''
        Return the first and second derivatives of psi w.r.t. time
        The derivative should be computed with self.grid_tmp, not self.grid
        '''
        raise NotImplementedError
                                  
    def step(self):
      self.set_age(self.age + self.dt)

      self.grid_tmp.psi[:] = self.grid.psi
      self.dgrid_tmp.psi[:] = self.dgrid.psi
                              
      k1, l1 = self.psi_dot()
      self.grid_tmp.psi[...] += k1*self.dt/2
      self.dgrid_tmp.psi[...] += l1*self.dt/2
                              
      k2, l2 = self.psi_dot()
      self.grid_tmp.psi[...] += k2*self.dt/2
      self.dgrid_tmp.psi[...] += l2*self.dt/2
                              
      k3, l3 = self.psi_dot()
      self.grid_tmp.psi[...] += k3*self.dt
      self.dgrid_tmp.psi[...] += l3*self.dt
                              
      k4, l4 = self.psi_dot()
                              
      self.grid.psi[...] += self.dt / 6 * (k1 + 2*k2 + 2*k3 + k4)
      self.dgrid.psi[...] += self.dt / 6 * (l1 + 2*l2 + 2*l3 + l4)



class FirstOrderRK4(TemporalEvolver[T]):
    '''
    Performs RK4 on a ComplexGrid object to solve a PDE that is
    first-order in time
    '''
    def __init__(self, grid: T, dt: float):
        super().__init__(grid, dt)
    
        self.grid = self.subject
        '''
        Here self.grid == self.subject
        '''
        self.grid_tmp = self.grid.copy()
        self.grid_tmp.initialize_fft()

        if not self.grid.fft_initialized():
            self.grid.initialize_fft()



    @abstractmethod
    def psi_dot(self) -> npt.NDArray:
        '''
        Return the first derivative of psi w.r.t. time
        The derivative should be computed with self.grid_tmp, not self.grid
        '''
        raise NotImplementedError
                                  
    def step(self):
      self.age += self.dt

      self.grid_tmp.psi[...] = self.grid.psi
                              
      k1= self.psi_dot()
      self.grid_tmp.psi[...] += k1*self.dt/2
                              
      k2= self.psi_dot()
      self.grid_tmp.psi[...] += k2*self.dt/2
                              
      k3= self.psi_dot()
      self.grid_tmp.psi[...] += k3*self.dt
                              
      k4= self.psi_dot()
                              
      self.grid.psi[...] += self.dt / 6 * (k1 + 2*k2 + 2*k3 + k4)










raise ImportError('Module is outdated and deprecated')
import numpy as np
from ._complex import ComplexGrid
from ._real import RealGrid



def load_grid(filepath: str, is_complex=False):
    state = np.load(filepath)
    return import_grid(state, is_complex=False)


def import_grid(state, is_complex=False):
    psi = state['psi']
    shape = psi.shape
    if is_complex: 
        grid = ComplexGridND(shape)
        grid.set_psi(psi)
        return grid 
    else:
        grid = RealGridND(shape)
        grid.set_psi(psi)
        return grid









from typing import Tuple
import numpy as np
import numpy.typing as npt

import warnings

from .dtypes import PrecisionLike, FloatLike, get_real_dtype


def generate_xk(
    L: FloatLike, N: int, *,
    center: bool=False, real: bool=False,
    precision: PrecisionLike = 'double'
) ->    Tuple[
        npt.NDArray[np.floating], npt.NDArray[np.floating], 
        np.floating, np.floating]:
    """
    Generate x and k for Fourier transform given system size (L) and number of
    samples (N)
    """
    dtype = get_real_dtype(precision)

    if not isinstance(L, dtype):
        warnings.warn(f'Precision mismatch: {dtype} and {type(L)}')
        L = dtype(L)

    dx = L / N
    dk = np.pi*2 / L

    M = N // 2

    if center:
        x = np.linspace(-L/2, L/2, N, endpoint=False, dtype=dtype)
    else:
        x = np.linspace(0, L, N, endpoint=False, dtype=dtype)

    if real:
        k = np.array([n*dk for n in range(M+1)], dtype=dtype)
    else:
        k = np.array([n*dk for n in range(0, M)] + [n*dk for n in range(M-N, 0)], dtype=dtype)

    return x, k, dx, dk






'''
Grid module. Grids are numpy array wrappers that support FFT operations without
length scales.
'''

from ._base import Grid

from ._complex import ComplexGrid
from ._real import RealGrid

from ._1d import ComplexGrid1D, RealGrid1D, Grid1D
from ._2d import ComplexGrid2D, RealGrid2D, Grid2D





from __future__ import annotations
from types import TracebackType
from typing import ContextManager, Protocol, Type


class EnterFunc(Protocol):
  def __call__(self): ...


class ExitFunc(Protocol):
  def __call__(self, __exc_type: Type[BaseException] | None,                                
          __exc_value: BaseException | None,                                
          __traceback: TracebackType | None) -> bool | None:
      ...


class Context(ContextManager):
  def __init__(self, _enter: EnterFunc|None=None, _exit: ExitFunc|None=None):
      self._enter = _enter
      self._exit = _exit

  def __enter__(self) -> None:
      if self._enter is not None: self._enter()

  def __exit__(self, __exc_type: Type[BaseException] | None,                                
          __exc_value: BaseException | None,                                
          __traceback: TracebackType | None) -> bool | None:

      if self._exit is not None: return self._exit(__exc_type, __exc_value, __traceback)                                





from __future__ import annotations

from typing import TypeVar
from ...core import FloatLike
from ..base import Evolver


T = TypeVar('T')


class TemporalEvolver(Evolver[T]):
    '''
    For evolvers associated with time steps
    '''
    def __init__(self, subject: T, dt: FloatLike):
        super().__init__(subject)
        self.dt = dt
        self.set_age(0)

    def set_age(self, age: FloatLike):
        self.age = age
        self.data['age'] = age





from __future__ import annotations
from abc import abstractmethod
from typing import TYPE_CHECKING, Optional, Sequence, Tuple, TypeVar, final

import numpy as np
import numpy.typing as npt

from ..core import get_real_dtype, PrecisionStr, NPFloat, SizeLike

from ..grids import Grid

if TYPE_CHECKING:
    from typing_extensions import Self

T = TypeVar('T', np.complexfloating, np.floating)

class Field(Grid[T]):
    """
    Base class for fields.
    """
    _size: npt.NDArray[np.floating]
    _R: npt.NDArray[np.floating]
    _dR: npt.NDArray[np.floating]

    _K: npt.NDArray[np.floating]
    _K2: npt.NDArray[np.floating]
    _K4: npt.NDArray[np.floating]
    _K6: npt.NDArray[np.floating]
    _dK: npt.NDArray[np.floating]

    _volume: NPFloat
    _dV: NPFloat

    def __init__(self,
            size: Sequence[np.floating|float] | npt.NDArray[np.floating],
            shape: Tuple[int, ...], *,
            precision: PrecisionStr = 'double',
            fft_axes: Optional[Tuple[int,...]] = None
        ):

        "call Grid contructor"
        Grid[T].__init__(
            self, shape,
            precision=precision, 
            fft_axes=fft_axes
        )
        self._init_coordinate_vars()
        self.set_size(size)

    @abstractmethod
    def _update_coordinate_vars(self) -> None:
        """
        Given size, update _R, _K, _dR, _dK
        """

    @final
    def _init_coordinate_vars(self):
        """
        Called by the Field constructure to initialize _size, _R, _K, _dR, _dK
        arrays to ZEROS
        """
        dtype = get_real_dtype(self._precision)
        self._size = np.zeros((self.rank,), dtype=dtype)
        self._R = np.zeros((self.rank, *self.shape), dtype=dtype)
        self._K = np.zeros((self.rank, *self.shape_k), dtype=dtype)
        self._dR = np.zeros((self.rank,), dtype=dtype)
        self._dK = np.zeros((self.rank,), dtype=dtype)
    
    def set_size(self, size: SizeLike):
        """
        Set system size (dimension lengths) and update size, r, k, dr, dk etc
        """
        self.validate_size(size)

        self._size[...] = size
        self._update_coordinate_vars()

        self._K2 = np.sum(self._K**2, axis=0)
        self._K4 = self._K2**2
        self._K6 = self._K2**3

        self._volume = np.prod(self._size)
        self._dV = np.prod(self._dR)

    def validate_size(self, size: SizeLike):
        if isinstance(size, np.ndarray):
            if size.ndim != 1:
                raise ValueError(f'ndarray with shape {size.shape} is not a valid field size')

            if not np.all(np.isreal(size)):
                raise ValueError(f'ndarray with dtype {size.dtype} is not a valid field size')
            
        if len(size) != self.rank:
            raise ValueError(f'size {size} is incompatible with current shape {self.shape}')

    @property
    def size(self): return self._size

    @property
    def r(self):
        'coordinates, shape = (rank, d1, d2, ..., dN)'
        return self._R

    @property
    def k(self):
        'k-space coordinates (frequencies), shape = (rank, d1, d2, ..., dN)'
        return self._K

    @property
    def k2(self):
        'k squared, shape = (d1, d2, ..., di/2+1, ... dN)'
        return self._K2

    @property
    def k4(self):
        'k^4, shape = (d1, d2, ..., di/2+1, ..., dN)'
        return self._K4

    @property
    def k6(self):
        'k^6, shape = (d1, d2, ..., di/2+1, ..., dN)'
        return self._K6

    @property
    def dr(self):
        'coordinate spacings, shape = (rank,)'
        return self._dR

    @property
    def dk(self):
        'k-space spacings, shape = (rank,)'
        return self._dK

    @property
    def volume(self) -> NPFloat:
        'real space volume'
        return self._volume

    @property
    def dv(self) -> NPFloat:
        'real space volume element'
        return self._dV

    @property
    def R(self): 'Deprecated; use self.r instead'; return self.r

    @property
    def K(self): 'Deprecated; use self.k instead'; return self.k

    @property
    def dR(self): 'Deprecated; use self.dr instead'; return self.dr

    @property
    def dK(self): 'Deprecated; use self.dk instead'; return self.dk

    @property
    def Volume(self): 'Deprecated; use self.volume instead'; return self.volume

    def copy(self) -> Self:
        field1 = self.__class__(
                self.size, self.shape,
                precision=self._precision,
                fft_axes=self._fft_axes
        )
        field1.psi[...] = self.psi
        return field1





from __future__ import annotations
from typing import TypeVar
from typing_extensions import Self

import numpy as np


from ..core import PrecisionStr

from ._base import Grid
from ._complex import ComplexGrid
from ._real import RealGrid


T = TypeVar('T', np.complexfloating, np.floating)

class Grid1D(Grid[T]):
    def __init__(self,
            n: int, *,
            precision: PrecisionStr = 'double'):
        super().__init__((n,), precision=precision, fft_axes=(0,))

    @property
    def n(self): 
        """=shape[0], the number of samples"""
        return self.shape[0]

    @property
    def N(self):
        """
        Deprecated; use self.n instead.
        """
        return self.n

    def copy(self) -> Self:
        g = self.__class__(self.n, precision=self._precision)
        g.psi[...] = self.psi
        return g

# class ComplexGrid1D(Grid1D[np.complexfloating], ComplexGrid):
class ComplexGrid1D(ComplexGrid, Grid1D[np.complexfloating]):
    """
    1D complex grid; fft axis is always axis 0.
    """


# class RealGrid1D(Grid1D[np.floating], RealGrid):
class RealGrid1D(RealGrid, Grid1D[np.floating]):
    """
    1D real grid; fft axis is always axis 0.
    """







from .consts import pi, e

from .dtypes import (PrecisionLike, PrecisionStr, 
                     NPFloat, NPComplex, FloatLike, ComplexLike,
                     SizeLike,
                     get_real_dtype, get_complex_dtype)

from .fourier import generate_xk

from .data import scalarize

from .generics import generic

from .version import __version__



from __future__ import annotations
from typing import Optional, Tuple, TypeVar
from typing_extensions import Self

import numpy as np

from ..core import PrecisionStr
from ._base import Grid
from ._complex import ComplexGrid
from ._real import RealGrid


T = TypeVar('T', np.complexfloating, np.floating)

class Grid2D(Grid[T]):
    def __init__(self, 
        nx: int, ny: int, *,
        precision: PrecisionStr = 'double', 
        fft_axes: Optional[Tuple[int,...]]=None):

        super().__init__(
                (nx, ny),
                precision=precision,
                fft_axes=fft_axes)

    @property
    def nx(self):
        """
        = shape[0], the number of samples in the x direction
        """
        return self.shape[0]

    @property
    def ny(self):
        """
        = shape[1], the number of samples in the y direction
        """
        return self.shape[1]

    def copy(self) -> Self:
        g = self.__class__(
                self.nx, self.ny,
                precision=self.precision,
                fft_axes=self.fft_axes)

        g.psi[...] = self.psi
        return g


# class ComplexGrid2D(Grid2D[np.complexfloating], ComplexGrid):
class ComplexGrid2D(ComplexGrid, Grid2D[np.complexfloating]):
    """
    2D complex grid
    """


# class RealGrid2D(Grid2D[np.floating], RealGrid):
class RealGrid2D(RealGrid, Grid2D[np.floating]):
    """
    2D real grid
    """




from typing import List, Sequence, Tuple
import numpy as np
import numpy.typing as npt



def prod(arr: Sequence[float]):
    """product of all the elements in an array"""
    r = 1
    for a in arr: r *= a
    return r


def gen_shape(rank: int, min_log_numel: float, max_log_numel: float):
    """Generate shape"""
    log_numel = np.random.rand() * (max_log_numel - min_log_numel) + min_log_numel

    seps = np.random.rand(rank-1).tolist()
    seps = [0] + sorted(seps) + [1]
    
    shape_ = []

    for i in range(rank):
        s = np.exp(log_numel * (seps[i+1] - seps[i]))
        shape_.append(int(s))

    return tuple(shape_)


def gen_shapes(n: int, ranks: Sequence[int],
               max_log_numel: float, min_log_numel: float) -> List[Tuple[int,...]]:
    """Generate multiple shapes"""
    res = []
    for _ in range(n):
        rank = np.random.choice(ranks)
        shape = gen_shape(rank, max_log_numel=max_log_numel, min_log_numel=min_log_numel)
        res.append(shape)
    return res


def gen_size(shape: Tuple[int,...], min, max):
    """
    Return a random size based on shape
    """
    rank = len(shape)
    size = []
    for _ in range(rank):
        size.append(np.random.rand()*(max-min) + min)
    return tuple(size)


def gen_random_sequence(length: int, min_prod: float, max_prod: float, *, integ: bool=False):
    """
    Genereate a sequence of fixed length such that min_prod < product of all elements < max_prod
    """
    min_log_prod = np.log(min_prod)
    max_log_prod = np.log(max_prod)

    log_prod = np.random.rand() * (max_log_prod - min_log_prod) + min_log_prod

    seps = np.random.rand(length-1).tolist()
    seps = [0] + sorted(seps) + [1]
    
    res = []
    for i in range(length):
        s = np.exp(log_prod * (seps[i+1] - seps[i]))
        if integ:
            s = round(s)
        res.append(s)
    return tuple(res)


def gen_fft_axes(rank: int):
    """Generate FFT axes"""
    n_ax = np.random.randint(1, rank+1)
    return tuple(np.random.choice(range(rank), replace=False, size=n_ax).tolist())


def float_tup_repr(tup, fmt: str='.3e'):
    return '(' + ','.join([f'{s:{fmt}}' for s in tup]) + ')'


def int_tup_repr(tup: Tuple[int,...]):
    return '(' + ','.join([str(s) for s in tup]) + ')'


def get_test_id(size, shape, fft_axes):
    """Test IDs"""
    return f'size={float_tup_repr(size)} shape={int_tup_repr(shape)} fft={int_tup_repr(fft_axes)}'


def gen_array(shape: Tuple[int,...], 
              min_offset: float = 0, max_offset: float = 0,
              min_scale: float = 1, max_scale: float = 1,
              complex: bool = False
              ):
    """
    Generate random array
    """
    if complex:
        a = gen_array(shape, min_offset, max_offset, min_scale, max_scale)
        b = gen_array(shape, min_offset, max_offset, min_scale, max_scale)
        return a + 1j*b

    offset = np.random.rand() * (max_offset - min_offset) + min_scale
    logscale =  np.random.rand() * (np.log(max_scale) - np.log(min_scale)) + np.log(min_scale)
    return (np.random.rand(*shape) + offset) * np.exp(logscale)


def adaptive_atol_allclose(f1: npt.ArrayLike, f2: npt.ArrayLike, rtol: float, *,
                           atol_factor: float = 1e-5):
    """
    Compare two arrays symmetrically
    """
    f1_ = np.asarray(f1)
    f2_ = np.asarray(f2)

    dev1 = np.sqrt(np.mean(np.abs(f1-np.mean(f1))**2))
    dev2 = np.sqrt(np.mean(np.abs(f1-np.mean(f1))**2))

    dev = np.sqrt((dev1**2+dev2**2)/2)

    atol = dev * atol_factor

    mag = np.sqrt((np.abs(f1_**2) + np.abs(f2_**2)) / 2)

    return np.all(np.abs(f1_ - f2_) <= rtol * mag + atol) # type: ignore


real_dtype_table = {
    'SINGLE': np.single,
    'DOUBLE': np.double,
    'LONGDOUBLE': np.longdouble
}


complex_dtype_table = {
    'SINGLE': np.csingle,
    'DOUBLE': np.cdouble,
    'LONGDOUBLE': np.clongdouble
}


nbytes_table = {'SINGLE': 4, 'DOUBLE': 8, 'LONGDOUBLE': 16}






'''
Field module. Fields are grids with length scales.
'''

from ._base import Field
from ._complex import ComplexField
from ._real import RealField

from ._1d import ComplexField1D, RealField1D, Field1D

from ._2d import ComplexField2D, RealField2D, Field2D







from .base import EvolverHooks, combine

from .default import DefaultHooks

from .display import Display, Progress, Panel, Text

from .earlystop import EarlyStopping, DetectSlow

from .monitor import MonitorValues

from .menu import Menu



from __future__ import annotations
from typing import Any, List, Literal, Tuple, TypeVar, Union, overload, Optional

import numpy as np
from matplotlib import pyplot as plt

from .fields import ComplexField2D, RealField2D
from .grids import ComplexGrid


def plot(fields: ComplexField2D|List[ComplexField2D],
        cmap: str='jet', show: bool=True, vlim: Tuple[float, float]=(-1., 1.), 
        colorbar: bool=True, ncols=4, fig_dims=(4, 4)):

    if not type(fields) in [tuple, list]:
        assert isinstance(fields, ComplexField2D)
        fields = [fields]

    assert isinstance(fields, list)
    nrows = (len(fields)-1) // ncols + 1

    if len(fields) < ncols:
        ncols = len(fields)

    fig = plt.figure(figsize=(fig_dims[0]*ncols, fig_dims[1]*nrows))
    cms = []
    axs = []
    for i, field in enumerate(fields):
        ax = plt.subplot(nrows, ncols, i+1)
        cm = ax.pcolormesh(field.X, field.Y, np.real(field.psi), cmap=cmap, vmin=vlim[0], vmax=vlim[1], shading='nearest')
        if colorbar:
            plt.colorbar(cm, ax=ax, orientation='horizontal')

        ax.set_aspect('equal', adjustable='box')
        cms.append(cm)
        cms.append(ax)

    if show:
       plt.show() 
    else:
       return fig, axs

@overload
def set_size(field: RealField2D, Lx: float, Ly: float) -> RealField2D: ...
@overload
def set_size(field: ComplexField2D, Lx: float, Ly: float) -> ComplexField2D: ...
@overload
def set_size(field: ComplexField2D, Lx: float, Ly: float, in_place: Literal[True]) -> None: ...
@overload
def set_size(field: RealField2D, Lx: float, Ly: float, in_place: Literal[False]) -> RealField2D: ...
@overload
def set_size(field: ComplexField2D, Lx: float, Ly: float, in_place: Literal[False]) -> ComplexField2D: ...

def set_size(field: ComplexField2D, Lx: float, Ly: float, in_place: bool=False) -> Union[None, ComplexField2D, RealField2D]:
    if in_place:
        field.set_size(Lx, Ly)
        return
    else:
        field1 = field.copy()
        field1.set_size(Lx, Ly)
        return field1


@overload
def change_resolution(field: RealField2D, Nx: int, Ny: int) -> RealField2D: ...
@overload
def change_resolution(field: ComplexField2D, Nx: int, Ny: int) -> ComplexField2D: ...
@overload
def change_resolution(field: ComplexField2D, Nx: int, Ny: int, in_place: Literal[True]) -> None: ...
@overload
def change_resolution(field: RealField2D, Nx: int, Ny: int, in_place: Literal[False]) -> RealField2D: ...
@overload
def change_resolution(field: ComplexField2D, Nx: int, Ny: int, in_place: Literal[False]) -> ComplexField2D: ...

def change_resolution(field: ComplexField2D, Nx: int, Ny: int, in_place: bool=False) -> Union[None, ComplexField2D, RealField2D]:
    psi1 = np.zeros((Nx, Ny))

    for i in range(Nx):
        for j in range(Ny):
            psi1[i,j] = field.psi[int(i*field.Nx/Nx),int(j*field.Ny/Ny)]

    if in_place:
        field.set_dimensions(field.Lx, field.Ly, Nx, Ny)
        field.set_psi(psi1)
        return
    else:
        field1 = field.copy()
        field1.set_dimensions(field.Lx, field.Ly, Nx, Ny)
        field1.set_psi(psi1)
        return field1


@overload
def extend(field: RealField2D, Mx: int, My: int) -> RealField2D: ...
@overload
def extend(field: ComplexField2D, Mx: int, My: int) -> ComplexField2D: ...
@overload
def extend(field: ComplexField2D, Mx: int, My: int, in_place: Literal[True]) -> None: ...
@overload
def extend(field: RealField2D, Mx: int, My: int, in_place: Literal[False]) -> RealField2D: ...
@overload
def extend(field: ComplexField2D, Mx: int, My: int, in_place: Literal[False]) -> ComplexField2D: ...

def extend(field: ComplexField2D, Mx: int, My: int, in_place: bool=False) -> None | ComplexField2D | RealField2D:
    Nx1 = field.Nx * Mx
    Ny1 = field.Ny * My

    psi1 = np.zeros((Nx1, Ny1))

    for i in range(Nx1):
        for j in range(Ny1):
            psi1[i,j] = field.psi[i%field.Nx,j%field.Ny]
    
    if in_place:
        field.set_dimensions(field.Lx*Mx, field.Ly*My, Nx1, Ny1)
        field.set_psi(psi1)
        return
    else:
        field1 = field.copy()
        field1.set_dimensions(field.Lx*Mx, field.Ly*My, Nx1, Ny1)
        field1.set_psi(psi1)
        return field1


@overload
def flip(field: RealField2D, axis: str) -> RealField2D: ...
@overload
def flip(field: ComplexField2D, axis: str) -> ComplexField2D: ...
@overload
def flip(field: ComplexField2D, axis: str, in_place: Literal[True]) -> None: ...
@overload
def flip(field: RealField2D, axis: str, in_place: Literal[False]) -> RealField2D: ...
@overload
def flip(field: ComplexField2D, axis: str, in_place: Literal[False]) -> ComplexField2D: ...

def flip(field: ComplexField2D, axis: str, in_place: bool=False) -> None | RealField2D | ComplexField2D:
    if axis not in ['X', 'Y']:
        raise ValueError(f'\'{axis}\' is not a valid axis for flipping') 

    psi1 = field.psi.copy()
    if axis == 'X':
        psi1 = psi1[:,::-1]
    if axis == 'Y':
        psi1 = psi1[::-1,:]

    if in_place:
        field.set_psi(psi1)
        return
    else:
        field1 = field.copy()
        field1.set_psi(psi1)
        return field1

@overload
def transpose(field: RealField2D) -> RealField2D: ...
@overload
def transpose(field: ComplexField2D) -> ComplexField2D: ...
@overload
def transpose(field: ComplexField2D, in_place: Literal[True]) -> None: ...
@overload
def transpose(field: RealField2D, in_place: Literal[False]) -> RealField2D: ...
@overload
def transpose(field: ComplexField2D, in_place: Literal[False]) -> ComplexField2D: ...

def transpose(field: ComplexField2D, in_place=False) -> None | RealField2D | ComplexField2D:
    psi1 = np.transpose(field.psi)
    if in_place:
        field.set_dimensions(field.Ly, field.Lx, field.Ny, field.Nx)
        field.set_psi(psi1)

    else:
        field1 = field.copy()
        field1.set_dimensions(field.Ly, field.Lx, field.Ny, field.Nx)
        field1.set_psi(psi1)
        return field1

@overload
def rotate(field: RealField2D, angle: str) -> RealField2D: ...
@overload
def rotate(field: ComplexField2D, angle: str) -> ComplexField2D: ...
@overload
def rotate(field: ComplexField2D, angle: str, in_place: Literal[True]) -> None: ...
@overload
def rotate(field: RealField2D, angle: str, in_place: Literal[False]) -> RealField2D: ...
@overload
def rotate(field: ComplexField2D, angle: str, in_place: Literal[False]) -> ComplexField2D: ...

def rotate(field: ComplexField2D, angle: str, in_place=False) -> None | RealField2D | ComplexField2D:
    if not angle in ['90', '180', '270']:
        raise ValueError(f'\'{angle}\' is not a valid angle for rotation')

    psi1 = field.psi.copy()
    if angle == '90':
        psi1 = np.transpose(psi1)[::-1,:]
    if angle == '180':
        psi1 = psi1[::-1,::-1]
    if angle == '270':
        psi1 = np.transpose(psi1)[:,::-1]

    if in_place:
        if angle in ['90', '270']:
            field.set_dimensions(field.Ly, field.Lx, field.Ny, field.Nx)
        field.set_psi(psi1)
        return
    else:
        field1 = field.copy()
        if angle in ['90', '270']:
            field1.set_dimensions(field.Ly, field.Lx, field.Ny, field.Nx)
        field1.set_psi(psi1)
        return field1

@overload
def crop(field: RealField2D, ratio_x1: float, ratio_x2: float, ratio_y1: float, ratio_y2: float) -> RealField2D: ...
@overload
def crop(field: ComplexField2D, ratio_x1: float, ratio_x2: float, ratio_y1: float, ratio_y2: float) -> ComplexField2D: ...
@overload
def crop(field: ComplexField2D, ratio_x1: float, ratio_x2: float, ratio_y1: float, ratio_y2: float, in_place: Literal[True]) -> None: ...
@overload
def crop(field: RealField2D, ratio_x1: float, ratio_x2: float, ratio_y1: float, ratio_y2: float, in_place: Literal[False]) -> RealField2D: ...
@overload
def crop(field: ComplexField2D, ratio_x1: float, ratio_x2: float, ratio_y1: float, ratio_y2: float, in_place: Literal[False]) -> ComplexField2D: ...

def crop(field: ComplexField2D, ratio_x1: float, ratio_x2: float, ratio_y1: float, ratio_y2: float, in_place=False):
    Nx2 = int(field.Nx * (ratio_x2 - ratio_x1))
    Ny2 = int(field.Ny * (ratio_y2 - ratio_y1))
    first_x = int(field.Nx * ratio_x1)
    first_y = int(field.Nx * ratio_y1)
    Lx2 = field.Lx*(ratio_x2-ratio_x1)
    Ly2 = field.Ly*(ratio_y2-ratio_y1)

    psi2 = field.psi[first_x:first_x+Nx2, first_y:first_y+Ny2]

    if in_place:
        field.set_dimensions(Lx2, Ly2, Nx2, Ny2)
        field.set_psi(psi2)
        return
    else:
        field1 = field.copy()
        field1.set_dimensions(Lx2, Ly2, Nx2, Ny2)
        field1.set_psi(psi2)
        return field1


def concat(field_a: ComplexField2D, field_b: ComplexField2D, 
        direction='horizontal', in_place=False):
    if not direction in ['horizontal', 'vertical']:
        raise ValueError(f'\'{direction}\' is not a valid direction')

    psi1 = None 
    if direction == 'horizontal':
        if field_a.Ny != field_b.Ny or not np.isclose(field_a.Ly, field_b.Ly, rtol=1e-9, atol=1e-10):
            raise FieldOperationError(f'could not horizontally concat two fields with different vertical dimensions:'+\
                    'Ly1={field_a.Ly}, Ny1={field_a.Ny}, Ly2={field_b.Ly}, Ny2={field_b.Ny}')

        psi1 = np.zeros(shape=(field_a.Nx+field_b.Nx, field_a.Ny)) 
        psi1[:field_a.Nx,:] = field_a.psi[:,:]
        psi1[field_a.Nx:,:] = field_b.psi[:,:]

        if in_place:
            field_a.set_dimensions(field_a.Lx+field_b.Lx, field_a.Ly, field_a.Nx+field_b.Nx, field_a.Ny)
            field_a.set_psi(psi1)
            return
        else:
            field1 = field_a.copy()
            field1.set_dimensions(field_a.Lx+field_b.Lx, field_a.Ly, field_a.Nx+field_b.Nx, field_a.Ny)
            field1.set_psi(psi1)
            return field1


    if direction == 'vertical':
        if field_a.Ny != field_b.Ny or not np.isclose(field_a.Ly, field_b.Ly, rtol=1e-9, atol=1e-10):
            raise FieldOperationError(f'could not vertically concat two fields with different horizontal dimensions:'+\
                    'Lx1={field_a.Lx}, Nx1={field_a.Nx}, Lx2={field_b.Lx}, Nx2={field_b.Nx}')

        psi1 = np.zeros(shape=(field_a.Nx, field_a.Ny+field_b.Ny)) 
        psi1[:,:field_a.Ny] = field_a.psi[:,:]
        psi1[:,field_a.Ny:] = field_b.psi[:,:]

        if in_place:
            field_a.set_dimensions(field_a.Lx, field_a.Ly+field_b.Ly, field_a.Nx, field_a.Ny+field_b.Ny)
            field_a.set_psi(psi1)
            return
        else:
            field1 = field_a.copy()
            field_a.set_dimensions(field_a.Lx, field_a.Ly+field_b.Ly, field_a.Nx, field_a.Ny+field_b.Ny)
            field_a.set_psi(psi1)
            return field_a


def insert(
        field_large: ComplexField2D, field_small: ComplexField2D, 
        position='center', in_place=False, position_format='ratio',
        periodic_boundary=False
    ):
    '''
    '''
    valid = False
    if position in ['left', 'right', 'top', 'bottom', 'center']:
        valid = True
    if type(position) is tuple:
        if len(position) == 2:
            valid = True
    if not valid:
        raise ValueError(f'\'{position}\' is not a valid insertion position')

    if not position_format in ['ratio', 'points', 'length']:
        raise ValueError(f'\'{position_format}\' is not a valid position format')

    if not (field_small.Lx <= field_large.Lx and field_small.Ly <= field_small.Ly):
        raise FieldOperationError('the first model must have smaller dimensions than the second one')

    box_Nx, box_Ny = int(field_small.Lx / field_large.Lx * field_large.Nx), int(field_small.Ly / field_large.Ly * field_large.Ny)
    psi_small_modified = change_resolution(field_small, box_Nx, box_Ny)
   
    box_X, box_Y = int(field_large.Nx/2 - box_Nx/2), int(field_large.Ny/2 - box_Ny/2)

    if position == 'left':
        box_X = 0

    if position == 'right':
        box_X = field_large.Nx - box_Nx

    if position == 'top':
        box_Y = 0
        
    if position == 'bottom':
        box_Y = field_large.Ny - box_Ny

    if type(position) is tuple:
        if position_format == 'ratio':
            box_X = int(position[0] * field_large.Nx)
            box_Y = int(position[1] * field_large.Ny)

        if position_format == 'points':
            box_X = int(position[0])
            box_Y = int(position[1])

        if position_format == 'length':
            box_X = int(position[0]/field_large.Lx * field_large.Nx)
            box_Y = int(position[1]/field_large.Ly * field_large.Ny)

        if (not periodic_boundary) and (box_X + box_Nx > field_large.Nx or box_Y + box_Ny > field_large.Ny):
            raise FieldOperationError('insertion out of boundary with periodic_boundary=False')

    psi1 = field_large.psi.copy()

    psi1[box_X:(box_X+box_Nx)%field_large.Nx,box_Y:(box_Y+box_Ny)%field_large.Ny] = psi_small_modified[:,:]

    if in_place:
        field_large.set_psi(psi1)
        return


    else:
        field1 = field_large.copy()
        field1.set_psi(psi1)
        return field1

@overload 
def interface(field_a: RealField2D, field_b: RealField2D, width:
        float, left=0.25, right=0.75) -> RealField2D: ...
@overload
def interface(field_a: ComplexField2D, field_b: ComplexField2D, width: float,
        left=0.25, right=0.75) -> ComplexField2D: ...
@overload
def interface(field_a: ComplexField2D, field_b: ComplexField2D, width: float,
        left: float=0.25, right: float=0.75, in_place: Literal[True]=True) -> None:
    ...
@overload
def interface(field_a: RealField2D, field_b: RealField2D, width: float,
        left: float=0.25, right: float=0.75, in_place: Literal[False]=False) -> RealField2D: ...
@overload
def interface(field_a: ComplexField2D, field_b: ComplexField2D, width: float,
        left: float=0.25, right: float=0.75, in_place: Literal[False]=False) -> ComplexField2D: ...

def interface(field_a: ComplexField2D, field_b: ComplexField2D, width: float,
        left: float=0.25, right: float=0.75, in_place: bool=False) -> None | ComplexField2D | RealField2D:
    try:
        assert np.isclose(field_a.Lx, field_b.Lx, rtol=1e-6, atol=1e-6)
        assert np.isclose(field_a.Ly, field_b.Ly, rtol=1e-6, atol=1e-6)
        assert field_a.Nx == field_b.Nx
        assert field_a.Ny == field_b.Ny

    except AssertionError as e:
        raise ValueError(f'two fields with different dimensions can not be combined using interface()')


    X, Y = field_a.X, field_a.Y

    xa = field_a.Lx * left
    xb = field_a.Lx * right

    bump = (1+np.tanh((X-xa)/width))/2 * (1+np.tanh((-X+xb)/width))/2

    psi1 = field_a.psi * bump + field_b.psi * (1-bump)

    if in_place:
        field_a.set_psi(psi1)
        return

    else:
        field1 = field_a.copy()
        field1.set_psi(psi1)
        return field1

T = TypeVar('T', bound=ComplexGrid)



@overload
def liquefy(field: T, psi0=None, *, in_place: Literal[False]=False) -> T: ...

@overload
def liquefy(field: ComplexGrid, psi0=None, *, in_place: Literal[True]) -> None: ...

@overload
def liquefy(field: T, psi0=None, *, in_place: bool=False) -> None|T: ...

def liquefy(
    field, psi0=None, *,
    in_place: bool=False):

    if psi0 is None:
        psi0 = np.mean(field.psi)
    
    if in_place:
        field.set_psi(psi0)
        return

    else:
        field1 = field.copy()
        field1.set_psi(psi0)
        return field1



class FieldOperationError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message





from __future__ import annotations

import numpy as np
import numpy.typing as npt


from ..core.fourier import generate_xk

from ..grids import RealGrid
from ._base import Field


class RealField(Field[np.floating], RealGrid):
    def _update_coordinate_vars(self) -> None:
        R, K, DR, DK = [], [], [], []
        for i in range(self.rank):
            if i == self.last_fft_axis:
                r, k, dr, dk = generate_xk(self.size[i], self.shape[i], 
                                           real=True, precision=self.precision)
            else:
                r, k, dr, dk = generate_xk(self.size[i], self.shape[i], 
                                           precision=self.precision) 
            R.append(r)
            K.append(k)
            DR.append(dr)
            DK.append(dk)

        self._R[...] = np.meshgrid(*R, indexing='ij')
        self._K[...] = np.meshgrid(*K, indexing='ij')
        self._dR[...] = np.array(DR)
        self._dK[...] = np.array(DK)




from __future__ import annotations
from typing import Tuple, TypeVar, overload, Optional
from typing_extensions import Self
import numpy as np
import numpy.typing as npt

from ..core import PrecisionStr, FloatLike

from ..grids import Grid2D
from ._complex import ComplexField
from ._real import RealField
from ._base import Field

T = TypeVar('T', np.complexfloating, np.floating)

class Field2D(Field[T], Grid2D):
    def __init__(self, 
            lx: FloatLike, ly: FloatLike,
            nx: int, ny: int, *,
            precision: PrecisionStr = 'double',
            fft_axes: Optional[Tuple[int,...]]=None
            ):

        super().__init__(
            (lx, ly), (nx, ny),
            precision=precision, fft_axes=fft_axes
        )

    @property
    def lx(self) -> np.floating:
        return self.size[0]

    @property
    def ly(self) -> np.floating:
        return self.size[1]

    @property
    def nx(self):
        return self.shape[0]

    @property
    def ny(self):
        return self.shape[1]

    @property
    def x(self) -> npt.NDArray[np.floating]:
        return self.r[0]

    @property
    def y(self) -> npt.NDArray[np.floating]:
        return self.r[1]

    @property
    def dx(self) -> np.floating:
        return self.dr[0]

    @property
    def dy(self) -> np.floating:
        return self.dr[1]

    @property 
    def kx(self) -> npt.NDArray[np.floating]:
        return self.k[0]

    @property 
    def ky(self) -> npt.NDArray[np.floating]:
        return self.k[1]

    @property
    def dkx(self) -> np.floating:
        return self.dk[0]

    @property
    def dky(self) -> np.floating:
        return self.dk[1]

    @overload
    def set_size(self, lx: FloatLike, ly: FloatLike, /): ...
    @overload
    def set_size(self, size: Tuple[FloatLike, FloatLike], /): ...

    def set_size(self, 
            x1: Tuple[FloatLike, FloatLike]|FloatLike, 
            x2: Optional[FloatLike]=None, /):

        if isinstance(x1, tuple): 
            assert x2 is None
            Field[T].set_size(self, x1)
        else:
            assert x2 is not None
            Field[T].set_size(self, (x1, x2))

    def copy(self) -> Self:
        f = self.__class__(
                self.lx, self.ly, 
                self.nx, self.ny,
                fft_axes=self.fft_axes,
                precision=self._precision
            )

        f.psi[...] = self.psi
        return f


class ComplexField2D(Field2D[np.complexfloating], ComplexField):
    """
    2D complex field
    """


class RealField2D(Field2D[np.floating], RealField):
    """
    2D real field
    """




from __future__ import annotations

from typing import Tuple, final

import numpy as np
import pyfftw


from ..core import get_complex_dtype

from ._base import Grid


class ComplexGrid(Grid[np.complexfloating]):
    """
    A ComplexGridND object is a complex array of shape (d1, d2, .., dN)
    equipped with fourier transform. No length scales are associated with the
    grid. 
    """
    @final
    def _init_grid_data(self, shape: Tuple[int, ...]):
        self._psi = pyfftw.zeros_aligned(shape, dtype=get_complex_dtype(self._precision))
        self._psi_k = pyfftw.zeros_aligned(shape, dtype=get_complex_dtype(self._precision))

    @property
    def isreal(self): return False

    # def set_psi(self, 
    #         psi1: Union[complex,
    #                     npt.NDArray[np.complexfloating],
    #                     npt.NDArray[np.floating]]
    #     ) -> None:
    #     """
    #     Set grid data.
    #
    #     Parameters: psi1: new grid data, can be either scalar (float) or
    #     np.ndarray. If a scalar is given, all entries are set to the given value.
    #     """
    #     if not np.isscalar(psi1):
    #         assert isinstance(psi1, np.ndarray)
    #         if psi1.shape != self.shape:
    #             raise ValueError(f'array has incompatible shape {psi1.shape} with {self.shape}')
    #
    #     self.psi[...] = psi1

    # def __init__(self, 
    #         shape: Tuple[int, ...], *,
    #         precision: PrecisionStr='double',
    #         fft_axes: Optional[Tuple[int,...]]=None
    #     ):
    #     """
    #     Set the resolution (i.e. shape).
    #     
    #     Parameters: shape: tuple of integers (d1, d2, ..., dN),
    #     """
    #     self._precision = precision
    #
    #     self._psi = pyfftw.zeros_aligned(shape, dtype=get_complex_dtype(self._precision))
    #     self._psi_k = pyfftw.zeros_aligned(shape, dtype=get_complex_dtype(self._precision))
    #
    #     if fft_axes is None:
    #         self._fft_axes = tuple(np.arange(self.rank))
    #     else:
    #         self._fft_axes = fft_axes


    # def save(self, fname: str, verbose=False) -> None:
    #     """Save the grid data into a file.
    #
    #     Parameters:
    #         fname: the base file name. A .grid extension will be appended.
    #         verbose: whether to print out details
    #     """
    #     tmp_name = f'{fname}.tmp.file'
    #     if verbose:
    #         self.yell(f'dumping profile data to {fname}.grid')
    #     np.savez(tmp_name, **self.export_state()) 
    #     shutil.move(f'{tmp_name}.npz', f'{fname}.grid')

    # def export_state(self) -> dict:
    #     """Export the grid state to a dictionary
    #     """
    #     state = {'psi': self.psi.copy()}
    #     return state



from __future__ import annotations
from abc import ABC, abstractmethod, abstractproperty

from typing import TYPE_CHECKING, Generic, Literal, Optional, Tuple, TypeVar, Union

import numpy as np
import numpy.typing as npt

import pyfftw

from ..core import PrecisionStr, FloatLike, generic

if TYPE_CHECKING:
    from typing_extensions import Self

T = TypeVar('T', np.complexfloating, np.floating)


@generic
class Grid(ABC, Generic[T]):
    """
    Base class for grids.
    The generic type T refers to the dtype of the wrapped numpy array (self.psi).
    """
    
    _fft: Union[pyfftw.FFTW, None] = None
    _ifft: Union[pyfftw.FFTW, None] = None

    _psi: npt.NDArray[T]
    _psi_k: npt.NDArray[np.complexfloating]

    _precision: PrecisionStr
    _fft_axes: Tuple[int, ...]

    def __init__(self, 
            shape: Tuple[int, ...], *,
            precision: PrecisionStr='double',
            fft_axes: Optional[Tuple[int,...]]=None
        ):

        self._precision = precision.upper() # type: ignore

        if fft_axes is None:
            self._fft_axes = tuple(np.arange(len(shape)))
        else:
            self._fft_axes = fft_axes

        self._init_grid_data(shape) # init psi and psi_k
        
    @abstractmethod
    def _init_grid_data(self, shape: Tuple[int,...]) -> None:
        """
        Initialize psi and psi_k
        """

    @abstractproperty
    def isreal(self) -> bool:
        """
        Return whether the field or grid object holds real data
        """
        ...

    def set_psi(self, psi1: npt.NDArray[T]|FloatLike) -> None:
        """
        Deprecated: use psi[...] = ... instead
        """
        raise RuntimeError('use psi[...] = ... to assign grid data instead')

    def initialize_fft(self, **fftwargs) -> None:
        """
        Initialize the FFTW forward and backward plans. By default the
        fourier transform plans are not initialized as it can take a
        considerable amount of time.

        Parameters:
            **fftwargs: keyword arguments to be passed to pyfftw.FFTW()

        """
        psi_tmp = self.psi.copy()
        self._fft = pyfftw.FFTW(
                self._psi, self._psi_k,
                direction='FFTW_FORWARD',
                axes=self._fft_axes, **fftwargs)

        self._ifft = pyfftw.FFTW(
                self._psi_k, self._psi,
                direction='FFTW_BACKWARD',
                axes=self._fft_axes, **fftwargs)

        # self.set_psi(psi_tmp)
        self.psi[...] = psi_tmp

    def fft(self):
        """
        Run forward FFT on psi and store results to psi_k
        If FFT plans are not initialized, a TypeError will be raised
        """
        self._fft() # type: ignore

    def ifft(self):
        """
        Run backward FFT on psi_k and store results to psi
        If FFT plans are not initialized, a TypeError will be raised
        """
        self._ifft() # type: ignore
    
    def fft_initialized(self) -> bool:
        """
        Return whether the FFTW plans are initialized.
        """
        return not (self._fft is None or self._ifft is None)
        
    @property
    def shape(self): 
        """
        Return the shape in real space (x-space / time domain etc)
        """
        return self.psi.shape

    @property
    def shape_k(self):
        """
        Return the shape in Fourier (k-space / frequency domain etc)
        """
        return self.psi_k.shape

    @property
    def rank(self): 
        """
        Return the number of axes.
        Same as numpy's ndim.
        """
        return len(self.shape)

    @property
    def numel(self) -> int: 
        """
        Return the number of elements.
        Same as numpy's size() or torch's numel()
        """
        return np.prod(self.shape)
    
    @property
    def fft_axes(self):
        """
        Return the axes along which FFT is performed
        """
        return self._fft_axes

    @property
    def last_fft_axis(self): 
        """
        Return the last FFT axis.
        Useful when dealing with RFFTs.
        """
        return self._fft_axes[-1]

    @property
    def psi(self):
        """
        Real space data
        To assign values, use .psi[...] = ..., not .psi = ...
        """
        return self._psi

    @property
    def psi_k(self): 
        """
        k-space data
        To assign values, use .psi_k[...] = ..., not .psi_k = ...
        """
        return self._psi_k

    @property
    def precision(self) -> Literal['SINGLE', 'DOUBLE', 'LONGDOUBLE']:
        """
        Return the floating point precision
        """
        return self._precision # type: ignore

    def copy(self) -> Self:
        """
        Build a new object with the same grid data.
        Must be overriden if __init__ signature is overriden
        """
        grid1 = self.__class__(
                    self.shape,
                    precision=self.precision,
                    fft_axes=self.fft_axes
                )
        grid1.psi[...] = self.psi
        return grid1






from __future__ import annotations
from typing import TYPE_CHECKING, Callable, Dict, Generic, List, Optional, TypeVar, final



from ...core import generic

if TYPE_CHECKING:
    from ..base import Evolver
    from typing_extensions import Concatenate, ParamSpec

    P = ParamSpec('P')
T = TypeVar('T')

__hook_methods: Dict[str, bool] = {}

def hook(meth: Callable[Concatenate[EvolverHooks, P], None]):
    __hook_methods[meth.__name__] = False
    return meth

def reverse_hook(meth: Callable[Concatenate[EvolverHooks, P], None]):
    __hook_methods[meth.__name__] = True
    return meth



@generic
class EvolverHooks(Generic[T]):

    evolver: Evolver[T]

    def on_start(self, n_steps: int, n_epochs: Optional[int]): ...

    def on_step(self, step: int): ...

    def on_end(self): ...

    def enter_(self): ...

    def exit_(self, *args): ...

    @hook
    def on_multisteps_start(self, n_steps: int, n_epochs: int):
        '''
        Called before multistep evolution starts.
        self.on_start() is called by default.
        '''
        self.on_start(n_steps, n_epochs)
    
    @hook
    def multisteps_enter(self):
        self.enter_()

    @hook
    def on_multisteps_step(self, step: int):
        '''
        Called every [n_steps] during multistep evolultion. self.on_step() is
        called by default.
        '''
        self.on_step(step)


    @reverse_hook
    def multisteps_exit(self, *args):
        self.exit_(*args)

    @reverse_hook
    def on_multisteps_end(self):
        '''
        Called after multistep evolution ends.
        self.on_end() is called by default. Call order is reversed.
        '''
        self.on_end()

    @hook
    def on_nonstop_start(self, n_steps: int):
        '''
        Called before nonstop evolution starts.
        self.on_start() is called by default.
        '''
        self.on_start(n_steps, None)

    @hook
    def nonstop_enter(self):
        self.enter_()

    @hook
    def on_nonstop_step(self, step: int):
        '''
        Called every [n_steps]. self.on_step() is called by default.
        '''
        self.on_step(step)

    @reverse_hook
    def nonstop_exit(self, *args):
        '''
        Call order is reversed
        '''
        self.exit_(*args)

    @reverse_hook
    def on_nonstop_end(self):
        '''
        Called after nonstop evolution ends.
        self.on_end() is called by default.
        Call order ir reversed.
        '''
        self.on_end()

    @hook
    def pre_interrupt(self):
        '''
        Called when evolver is interrupted and **before**
        the field lock is acquired
        '''

    @hook
    def on_interrupt(self):
        '''
        Called when evolver is interrupted and **after**
        the field lock is acquired.
        '''

    @reverse_hook
    def post_interrupt(self):
        '''
        Called after evolver is interrupted
        '''


    @hook
    def bind(self, evolver: Evolver[T]):
        '''
        Bind evolver to hooks
        '''
        self.evolver = evolver

    def __add__(self, other: EvolverHooks):  
        return combine(self, other)



def combine(*hooks: EvolverHooks):
    '''
    Combine various hooks
    '''

    eh = EvolverHooks()

    def get_combined_func(name, reverse=False):
        if not reverse:
            def _combined(*args, **kwargs):
                handled = None
                for h in hooks:
                    handled_ = getattr(h, name)(*args, **kwargs)
                    handled = none_or(handled, handled_)
                return handled
            return _combined
        else:
            def _combined(*args, **kwargs):
                handled = None
                for h in hooks[::-1]:
                    handled_ = getattr(h, name)(*args, **kwargs)
                    handled = none_or(handled, handled_)
                return handled

            return _combined


    for methname, reverse in __hook_methods.items():
        setattr(eh, methname, get_combined_func(methname, reverse=reverse))

    return eh



def none_or(x: bool|None, y: bool|None):
    '''
        True, True/False/None -> True
        False, False/None -> False
        None, None -> None
    '''
    if x is None: return y
    if x is True: return True
    if x is False: return y is True





from .core import *
from .core import __version__

from .fields import *

from .grids import *

from . import proxies

from . import dynamics


from .field_util import (
        plot, set_size, liquefy, interface, insert,
        change_resolution, extend, flip, crop, rotate, transpose,
        concat
)



from __future__ import annotations


from typing import TYPE_CHECKING, Callable
import rich
from rich.live import Live

from torusgrid.dynamics.hooks.base import EvolverHooks

if TYPE_CHECKING:
    from torusgrid.dynamics.base import Evolver


class Menu(EvolverHooks):
    def __init__(self, 
            prompt: str, 
            action: Callable[[str, Evolver], None]
        ) -> None:
        self.prompt = prompt
        self.action = action

    def on_interrupt(self):

        #rich.get_console().log('Interrupted')
        live: Live = self.evolver.data['__live__']
        live.console.log('Interrupted')

        s = live.console.input(self.prompt)
        self.action(s, self.evolver)
        return True



from __future__ import annotations

from typing import TYPE_CHECKING, Tuple, TypeVar, overload

import numpy as np
import numpy.typing as npt

from ..core import PrecisionLike, FloatLike

from ..grids import Grid1D
from ._complex import ComplexField

from ._real import RealField
from ._base import Field

if TYPE_CHECKING:
    from typing_extensions import Self

T = TypeVar('T', np.complexfloating, np.floating)

class Field1D(Field[T], Grid1D):
    def __init__(
            self, 
            l: FloatLike, n: int, *,
            precision: PrecisionStr = 'double'):

        super().__init__(
            (l,), (n,),
            precision=precision, fft_axes=(0,)
        )

    @property
    def l(self) -> np.floating:
        return self.size[0]

    @property
    def n(self):
        return self.shape[0]

    @property
    def x(self) -> npt.NDArray[np.floating]:
        return self.r[0]

    @property
    def dx(self) -> np.floating:
        return self.dr[0]

    @property 
    def kx(self) -> npt.NDArray[np.floating]:
        return self.k[0]

    @property
    def dkx(self) -> np.floating:
        return self.dk[0]

    @overload
    def set_size(self, l: FloatLike, /): ...
    
    @overload
    def set_size(self, size: Tuple[FloatLike], /): ...

    def set_size(self, x: Tuple[FloatLike]|FloatLike, /):
        if isinstance(x, tuple): 
            Field[T].set_size(self, x)

        else:
            Field[T].set_size(self, (x,))

    def copy(self) -> Self:
        f = self.__class__(
                self.l, self.n,
                precision=self._precision
            )

        f.psi[...] = self.psi
        return f


#class ComplexField1D(Field1D[np.complexfloating], ComplexField):
class ComplexField1D(ComplexField, Field1D[np.complexfloating]):
    """
    1D complex field
    """

# class RealField1D(Field1D[np.floating], RealField):
class RealField1D(RealField, Field1D[np.floating]):
    """
    1D real field
    """




from __future__ import annotations
from rich.console import Console
from rich import get_console
from rich.style import StyleType
from rich.progress import (
        ProgressColumn, Progress, SpinnerColumn, MofNCompleteColumn,
        TimeElapsedColumn, TaskProgressColumn, TimeRemainingColumn, BarColumn
        )
from typing import Optional, Tuple


def make_progress_bar(console: Optional[Console]=None,
          bar_width_multiplier=.5,
          back_style: StyleType='bar.back',
          complete_style: StyleType='bar.complete',
          finished_style: StyleType='bar.finished',
          pulse_style: StyleType='bar.pulse',
  
          spinner: Optional[str|Tuple[str,str]]='dots', text: Optional[str]=None,
          count: bool=True, percentage: bool=True,
          time_remaining: bool=True, time_elapsed: bool=True, 
          *extra_columns: ProgressColumn,
          transient: bool=False
          ) -> Progress:
      '''
      Return a Rich Progress object
      '''
      if console is None:
          console = get_console()
  
      columns = []
  
      if spinner: 
          if isinstance(spinner, str):
              columns.append(SpinnerColumn(spinner))
          else:
              columns.append(SpinnerColumn(spinner[0], finished_text=spinner[1]))
  
      if text:
          columns.append(text)
  
      if count:
          columns.append(MofNCompleteColumn())
  
      columns.append(BarColumn(bar_width=int(bar_width_multiplier*console.width),
          style=back_style, complete_style=complete_style, finished_style=finished_style,
          pulse_style=pulse_style
          ))
      BarColumn()
  
      if percentage:
          columns.append(TaskProgressColumn())
  
      if time_remaining:
          columns.append(TimeRemainingColumn())
  
      if time_elapsed:
          columns.append(TimeElapsedColumn())
  
  
      progress = Progress(*columns, console=console, transient=transient)
  
      return progress




from __future__ import annotations
from abc import abstractmethod
from typing import List, Protocol, TypeVar


from ...core import FloatLike
from .temporal import TemporalEvolver
from ...grids import Grid


T = TypeVar('T', bound=Grid)


class Step(Protocol):
    def __call__(self, dt: FloatLike) -> None: ...


class SplitStep(TemporalEvolver[T]):
    def __init__(self, grid: T, dt: FloatLike):
        super().__init__(grid, dt)
        self.grid = self.subject
        '''
        Here self.grid == self.subject
        '''

        self.realspace_steps = self.get_realspace_steps()
        self.kspace_steps = self.get_kspace_steps()

    @abstractmethod
    def get_realspace_steps(self) -> List[Step]:
        raise NotImplementedError

    @abstractmethod
    def get_kspace_steps(self) -> List[Step]:
        raise NotImplementedError

    def step(self):
        self.set_age(self.age + self.dt)

        for step in self.realspace_steps:
            step(self.dt/2)
     
        self.grid.fft()

        for step in self.kspace_steps[:-1]:
            step(self.dt/2)

        self.kspace_steps[-1](self.dt)

        for step in self.kspace_steps[:-1][::-1]:
            step(self.dt/2)

        self.grid.ifft()

        for step in self.realspace_steps[::-1]:
            step(self.dt/2)







from typing import Optional, TypeVar
import rich

from rich.align import AlignMethod

from torusgrid.dynamics.hooks.default import DefaultHooks

from .base import EvolverHooks

from ...misc import console


from rich.live import Live
from rich.panel import Panel as Panel_
from rich.text import Text as Text_
from rich.console import Group
from rich import box

T = TypeVar('T')


class Display(EvolverHooks):
    '''
    Provide evolver.data with a live diplay object that can be accessed via
    evolver.data['__live__']

    '''
    def __init__(self): ...
    
    def on_start(self, n_steps, n_epochs):
        self.live = Live('', console=rich.get_console())
        
        if '__live__' in self.evolver.data.keys():
            raise RuntimeError('Only one Display object is allowed')

        self.evolver.data['__live__'] = self.live
        
    def enter_(self):
        self.live.start()
        #self.live.__enter__()
        
    def exit_(self, *args):
        #self.live.__exit__(*args)
        self.live.stop()
        del self.evolver.data['__live__']
        del self.live

    def pre_interrupt(self):
        ...

    def on_interrupt(self):
        ...

    def post_interrupt(self):
        ...
        # self.live = Live('', console=rich.get_console())
        # self.evolver.data['__live__'] = self.live
        # self.live.__enter__()
        #

class Panel(EvolverHooks):
    def __init__(self, title: str='', title_align: AlignMethod='center', **kwargs) -> None:
        self.title = title
        self.title_align: AlignMethod = title_align
        self.kwargs = kwargs

    def get_title(self):
        return self.title.format(**self.evolver.data)

    def on_start(self, 
            n_steps: int, n_epochs: Optional[int]):

        self.live: Live = self.evolver.data['__live__']
        self.panel = Panel_(
                Group(), 
                title=self.get_title(),
                title_align=self.title_align,
                **self.kwargs
            )
        self.evolver.data['__panel__'] = self.panel

    def enter_(self):
        self.live.update(self.panel)

    def exit_(self, *args):
        del self.evolver.data['__panel__']
        del self.panel


class Progress(EvolverHooks):
    '''
    '''
    def __init__(self, pbar_text: str, *, period: int=1):
        self.pbar_text = pbar_text
        self.period = period
    
    def on_start(self, n_steps, n_epochs):
        self.panel: Panel_ = self.evolver.data['__panel__']
        self.pbar = console.make_progress_bar(text='{task.fields[info]}')
        self.task = self.pbar.add_task('', total=n_epochs, start=(n_epochs is not None), info='')
        
    def enter_(self):
        self.panel.renderable = Group(self.panel.renderable, self.pbar)
        
    def exit_(self, *args):
        ...
    
    def on_step(self, step):
        self.pbar.advance(self.task)
        if step % self.period == 0:
            self.pbar.update(self.task, info=self.pbar_text.format(**self.evolver.data))



class Text(EvolverHooks):
    '''

    '''
    def __init__(self, text: str, *, period: int=1) -> None:
        self._text = text
        self.period = period

    def get_text(self):
        return self._text.format(**self.evolver.data)

    def on_start(self, n_steps: int, n_epochs: Optional[int]):
        self.panel: Panel_ = self.evolver.data['__panel__']
        self.textpanel = Panel_('', box=box.MINIMAL)
   
    def enter_(self):
        self.panel.renderable = Group(self.panel.renderable, self.textpanel)

    def on_step(self, step: int):
        if step % self.period == 0:
            self.textpanel.renderable = self.get_text()
    




from typing import TYPE_CHECKING, TypeVar, Type, Generic

T = TypeVar('T')

if TYPE_CHECKING:
    def generic(cls: Type[T]) -> Type[T]: 
        """
        Remove generic class inheritance and add trivial subscript
        such that A[T] = A.
        """
        return cls
else: 
    def generic(cls: type) -> type:
        """
        Remove generic class inheritance and add trivial subscript
        such that A[T] = A.
        """

        cls_ = add_trivial_subscript(remove_generic_base(cls))
        
        return cls_


def remove_generic_base(cls: type) -> type:
    """
    Remove Generic[T] base
    """
    metatype = type(cls)
    
    old_bases = cls.__bases__
    new_bases = list(cls.__bases__)

    for i, base in enumerate(old_bases):
        if base is Generic:
            del new_bases[i]

    if len(new_bases) == 0: new_bases = [object]

    cls_ = metatype(cls.__name__, tuple(new_bases), dict(cls.__dict__)) # type: ignore

    cls_.__name__ = cls.__name__
    cls_.__qualname__ = cls.__qualname__

    return cls_


def add_trivial_subscript(cls: type) -> type:
    """
    Add trivial class subscript
    """
    class cls_(cls): # type: ignore
        def __class_getitem__(cls, _):
            return cls
    cls_.__name__ = cls.__name__
    cls_.__qualname__ = cls.__qualname__
    return cls_




from __future__ import annotations
from numbers import Number
from typing import Any, Union, overload
import numpy as np


@overload
def scalarize(data: list) -> list: ...

@overload
def scalarize(data: dict) -> dict: ...

@overload
def scalarize(data: np.ndarray) -> np.ndarray | Number | np.generic: ...

@overload
def scalarize(data: str) -> str: ...

@overload
def scalarize(data: Number | np.generic) -> Number | np.generic: ...

def scalarize(data: dict | list | np.generic | Number | str | np.ndarray) -> Any:
    """
    When an NPZ file is loaded with np.load(), scalar values are stored as 0d
    arrays. This function converts all 0d arrays in a given object into
    scalars.
    """

    if np.isscalar(data):
        return data

    if isinstance(data, np.ndarray):
        try:
            return scalarize(data.item())
        except ValueError:
            return data

    if isinstance(data, dict):
        r = dict()
        for key, value in data.items():
            r[key] = scalarize(value)
        return r
    
    if isinstance(data, list):
        return [scalarize(item) for item in data]






raise ImportError('the history module is currently not maintained')

from michael960lib.common import IllegalActionError, ModifyingReadOnlyObjectError


class EvolverHistory:
    """
    A history object
    """
    def __init__(self):
        self.evolver_states = []
        self.state_functions = []
        self.committed = False 
        self.final_field_state = None

    def append_state_function(self, evolver_state: dict, sf: StateFunction):
        if self.committed:
            raise ModifyingReadOnlyObjectError(
            f'history object is already committed and hence not editable', self)

        self.state_functions.append(sf)

    def commit(self, label:str, field: ComplexField2D):
        if self.committed:
            raise IllegalActionError(f'history object (label=\'{self.label}\') is already committed')

        self.final_field_state = field.export_state()
        self.label = label
        self.committed = True 

    def is_committed(self):
        return self.committed

    def get_state_functions(self) -> List[StateFunction]:
        return self.state_functions

    def get_final_field_state(self):
        if not self.committed:
            raise IllegalActionError('cannot get final state from uncommitted evolver history')
        return self.final_field_state

    def get_label(self):
        return self.label

    def export(self) -> dict:
        if self.committed:
            state = dict()
            state['label'] = self.label
            state['state_functions'] = [sf.export() for sf in self.state_functions]
            return state
        else:
            raise IllegalActionError(
            'history object (label=\'{self.label}\') has not been committed and is therefore not ready to be exported')




from __future__ import annotations
import numpy as np

from .dtypes import PrecisionLike


PI_LONG = np.longdouble('3.1415926535897932384626433832795028841971693993')
PI_DOUBLE = np.double(PI_LONG)
PI_SINGLE = np.single(PI_LONG)


_pi = {
    'LONGDOUBLE': PI_LONG,
    'DOUBLE': PI_DOUBLE,
    'SINGLE': PI_SINGLE
}

def pi(precision: PrecisionLike) -> np.longdouble | np.double | np.single:
    return _pi[str(precision).upper()]


E_LONG = np.longdouble('2.718281828459045235360287471352662497757')
E_DOUBLE = np.double(E_LONG)
E_SINGLE = np.single(E_LONG)

_e = {
    'LONGDOUBLE': E_LONG,
    'DOUBLE': E_DOUBLE,
    'SINGLE': E_SINGLE
}

def e(precision: PrecisionLike) -> np.longdouble | np.double | np.single:
    return _e[str(precision).upper()]


